from time import sleep

import cv2
from fastapi import FastAPI
from fastapi.responses import StreamingResponse
import uvicorn

from guardcam.detection import Camera


def frame_to_content(frame, ext='png'):
    frame_bytes = cv2.imencode(f'.{ext}', frame)[1].tobytes()
    content_str = f'--frame\r\nContent-Type:image/{ext}\r\n\r\n'
    content = content_str.encode() + frame_bytes + b"\r\n"
    return content


class VideoStreamingServer:

    def __init__(self, host='0.0.0.0', port=7777):

        self.app = FastAPI()
        self.host = host
        self.port = port

    @staticmethod
    async def _cvcam_stream_content_generator():

        cam = cv2.VideoCapture(0)
        while True:
            ret, frame = cam.read()
            content = frame_to_content(frame, ext='jpg')
            yield content

    @staticmethod
    async def _camera_stream_content_generator(camera_instance):

        spf = 1 / camera_instance.fps
        while True:
            content = frame_to_content(camera_instance.current_frame, ext='jpg')
            yield content
            sleep(spf)

    def start(self, camera_instance=None):

        if camera_instance is not None:
            assert isinstance(camera_instance, Camera)
            frame_stream_generator = self._camera_stream_content_generator(camera_instance)
        else:
            frame_stream_generator = self._cvcam_stream_content_generator()

        @self.app.get('/')
        async def stream():
            return StreamingResponse(frame_stream_generator, media_type="multipart/x-mixed-replace; boundary=frame")

        print(f'\n\nUvicorn server running on http://{self.host}:{self.port}\n\n')
        uvicorn.run(app=self.app, host=self.host, port=self.port)


if __name__ == '__main__':

    vs = VideoStreamingServer()
    vs.start()




import os
import json
import requests
import logging

import cv2
import numpy as np

logging.basicConfig(format='%(asctime)s:%(name):½(level): %(message)s', datefmt='%d-%m-%y %H:%M:%S')


class TelegramBot:

    def __init__(self, bot_credentials):
        """

        :param bot_credentials: dict object containing bot credentials or file path to a credentials json file
        :type bot_credentials: dict or str
        """
        if isinstance(bot_credentials, str):
            with open(bot_credentials, 'r') as fp:
                bot_credentials = json.load(fp)

        self.name = bot_credentials.get('name', 'Bot')
        self.token = bot_credentials['token']
        self.chat_id = bot_credentials['chat_id']
        self.base_url = f'https://api.telegram.org/bot{self.token}'

    def set_chat_id(self, chat_id):
        self.chat_id = chat_id

    def set_token(self, token):
        self.token = token
        self.base_url = f'https://api.telegram.org/bot{self.token}'

    def send_message(self, message):
        url = f'{self.base_url}/sendMessage'
        data = {'chat_id': self.chat_id, 'text': f'[{self.name}] {message}'}

        response = requests.post(url=url, data=data)
        if response.status_code != 200:
            logging.error(f'Send message returned status code = {response.status_code}')
            return False

        return True

    def send_image(self, image_input, enc_ext='.jpg'):

        if isinstance(image_input, bytes):
            input_type = 'bytes'
            image_bytes = image_input

        elif isinstance(image_input, str):
            input_type = 'file'
            if os.path.isfile(image_input):
                image_bytes = open(image_input, 'rb')
            else:
                logging.error(f'Input image file: {image_input} not found')
                return False

        elif isinstance(image_input, np.ndarray):
            input_type = 'array'
            image_bytes = cv2.imencode(enc_ext, img=image_input)[1].tobytes()

        else:
            input_type = type(image_input)
            logging.error(f'Input image type {input_type} is not supported')
            return False

        url = f'{self.base_url}/sendPhoto'
        data = {'chat_id': self.chat_id}
        files = {'photo': image_bytes}
        response = requests.post(url=url, data=data, files=files)
        if response.status_code != 200:
            logging.error(f'Send photo returned status code = {response.status_code}')
            return False

        if input_type == 'file':
            image_bytes.close()

        return True

    def send_video(self, video_input):
        if isinstance(video_input, bytes):
            input_type = 'bytes'
            video_bytes = video_input

        elif isinstance(video_input, str):
            input_type = 'file'
            if os.path.isfile(video_input):
                video_bytes = open(video_input, 'rb')
            else:
                logging.error(f'Input video file: {video_input} not found')
                return False

        else:
            input_type = type(video_input)
            logging.error(f'Input video type {input_type} is not supported')
            return False

        url = f'{self.base_url}/sendVideo'
        data = {'chat_id': self.chat_id}
        files = {'video': video_bytes}
        response = requests.post(url=url, data=data, files=files)
        if response.status_code != 200:
            logging.error(f'Send video returned status code = {response.status_code}')
            return False

        if input_type == 'file':
            video_bytes.close()

        return True

    def send_file(self, file_input):
        if isinstance(file_input, bytes):
            input_type = 'bytes'
            file_bytes = file_input

        elif isinstance(file_input, str):
            input_type = 'file'
            if os.path.isfile(file_input):
                file_bytes = open(file_input, 'rb')

            else:
                logging.error(f'Input file: {file_input} not found')
                return False

        else:
            input_type = type(file_input)
            logging.error(f'Input file type {input_type} is not supported')
            return False

        url = f'{self.base_url}/sendDocument'
        data = {'chat_id': self.chat_id}
        files = {'document': file_bytes}
        response = requests.post(url=url, data=data, files=files)
        if response.status_code != 200:
            logging.error(f'Send file returned status code = {response.status_code}')
            return False

        if input_type == 'file':
            file_bytes.close()

        return True


if __name__ == '__main__':

    cwd = os.path.dirname(os.path.realpath(__file__))
    bot_credentials_file = os.path.join(cwd, 'credentials.json')
    with open(bot_credentials_file, 'r') as fp:
        credentials = json.load(fp)

    bot = TelegramBot(bot_credentials=credentials)
    bot.send_message('Example message')





import os
from threading import Thread
from time import time, sleep
from datetime import datetime

import cv2
import numpy as np


class Camera:
    """
    Class to run a webcam recording in a background thread. Current frame can be accessed through self.current_frame attribute
    """

    def __init__(self):
        self.current_frame = None
        self.cap = cv2.VideoCapture(0)
        self.fps = self.cap.get(cv2.CAP_PROP_FPS)

    def read(self):
        return True, self.current_frame

    def start(self):
        while True:
            ret, frame = self.cap.read()
            self.current_frame = frame

    def start_on_background(self):
        proc = Thread(target=self.start, daemon=True)
        proc.start()
        sleep(3)
        return proc


class VisualMovementDetector:

    def __init__(self, out_detections_dir, n_avg_frames, min_recording_time_after_detection=5, blur_kernel_size=5, min_cnt_area=700, out_fps=20, notif_bot=None):

        self.n_avg_frames = n_avg_frames
        self.avg_factor = 1.0 / self.n_avg_frames
        self.blur_kernel_size = blur_kernel_size
        self.min_cnt_area = min_cnt_area
        self.date = datetime.now().strftime("%d-%m-%Y")
        self.out_detections_dir = os.path.join(out_detections_dir, self.date)
        os.makedirs(self.out_detections_dir, exist_ok=True)
        self.last_detection_time = 0
        self.min_recording_time_after_detection = min_recording_time_after_detection
        self.out_fps = float(out_fps)
        self.notif_bot = notif_bot
        self.current_frame = np.zeros((256, 256), dtype=np.uint8)

    def _set_bg(self, frame):
        self.bg = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

    def add_frame_to_buffer(self, frame):
        frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        self.bg = cv2.addWeighted(self.bg, 1 - self.avg_factor, frame_gray, self.avg_factor, gamma=0)

    def evaluate_frame(self, frame):

        if self.bg is None:
            self._set_bg(frame)

        frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        proc_bg = cv2.GaussianBlur(self.bg, ksize=(self.blur_kernel_size, self.blur_kernel_size), sigmaX=0).astype(float)
        proc_frame = cv2.GaussianBlur(frame_gray, ksize=(self.blur_kernel_size, self.blur_kernel_size), sigmaX=0).astype(float)

        diff = np.abs(proc_bg - proc_frame)

        disp_diff = diff.astype(np.uint8)
        thresh = cv2.threshold(disp_diff, 25, 255, cv2.THRESH_BINARY)[1]

        cnts, _ = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        detection = False

        for c in cnts:
            if cv2.contourArea(c) > self.min_cnt_area:
                detection = True
                self.last_detection_time = time()
                break

        return detection

    def start(self, camera_instance: Camera):

        frame = camera_instance.current_frame
        self._set_bg(frame)

        video_writer = cv2.VideoWriter()
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        h, w = frame.shape[0:2]

        out_video_path = None
        while True:
            frame = camera_instance.current_frame
            detection = self.evaluate_frame(frame)
            curr_t = time()
            last_det_elapsed = curr_t - self.last_detection_time
            if detection or last_det_elapsed < self.min_recording_time_after_detection:

                # cv2.imshow('Movement detected!', frame)  # for debugging
                # cv2.waitKey(30)
                if not video_writer.isOpened():
                    det_time = datetime.now().strftime("%H:%M:%S")
                    msg = f'Intruder detected!! Time:  {self.date}_{det_time}'
                    print(msg)
                    if self.notif_bot is not None:
                        self.notif_bot.send_message(msg)

                    out_video_path = os.path.join(self.out_detections_dir, f'det_{det_time}.mp4')
                    video_writer.open(filename=out_video_path, fourcc=fourcc, fps=self.out_fps, frameSize=(w, h))

                video_writer.write(frame)

            else:
                cv2.destroyAllWindows()
                video_writer.release()
                if self.notif_bot is not None and out_video_path is not None:
                    self.notif_bot.send_video(out_video_path)
                    out_video_path = None

            self.add_frame_to_buffer(frame)



import argparse
from threading import Thread

from guardcam.detection import VisualMovementDetector
from guardcam.streaming import Camera, VideoStreamingServer
from guardcam.notification import TelegramBot


def parse_args():
    ap = argparse.ArgumentParser()
    ap.add_argument('-o', '--outDir', type=str, required=True, help='Output directory to store the detections captures')
    ap.add_argument('-nF', '--numAvgFrames', type=int, required=False, default=2, help='Number of averaging frames')
    ap.add_argument('-bS', '--blurKernelSize', type=int, required=False, default=7, help='Blur kernel size')
    ap.add_argument('-rT', '--recordTime', type=int, required=False, default=3, help='Minimum recording time after detection in seconds')
    ap.add_argument('-fps', '--outFPS', type=int, required=False, default=20, help='FPS for output detection captures')

    ap.add_argument('--streamingPort', type=int, required=False, default=None, help='Port for live video streaming. If None, no video streaming is served')
    ap.add_argument('--notificationTelegramBotCredentials', type=str, required=False, default=None, help='Bot credentials file to send notifications through Telegram. If None, disable notifications')
    return ap.parse_args()


def main():

    args = parse_args()

    bot_credentials = args.notificationTelegramBotCredentials
    notif_bot = TelegramBot(bot_credentials=bot_credentials) if bot_credentials is not None else None

    # start camera recording in background thread
    camera = Camera()
    cam_thread = camera.start_on_background()

    vmd = VisualMovementDetector(out_detections_dir=args.outDir,
                                 n_avg_frames=args.numAvgFrames,
                                 min_recording_time_after_detection=args.recordTime,
                                 blur_kernel_size=args.blurKernelSize,
                                 notif_bot=notif_bot)

    if args.streamingPort is None:
        vmd.start(camera_instance=camera)
    else:
        vmd_thread = Thread(target=vmd.start, kwargs={'camera_instance': camera}, daemon=True)
        vmd_thread.start()

        vss = VideoStreamingServer(port=args.streamingPort)
        vss.start(camera_instance=camera)


if __name__ == '__main__':
    main()



import os
import setuptools


def get_requirements():
    cwd = os.path.dirname(os.path.realpath(__file__))
    req_file = os.path.join(cwd, 'requirements.txt')
    with open(req_file, 'r') as fp:
        reqs = fp.read().splitlines()

    for i, pkg in enumerate(reqs):
        if '-e ' in pkg:
            git_link = pkg.split()[-1]
            pkg_name = pkg.split('=')[-1]
            reqs[i] = f'{pkg_name} @ {git_link}'

    return reqs


def get_version():
    cwd = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(cwd, 'guardcam', '__init__.py'), 'r') as fp:
        lines = fp.read().splitlines()

    for line in lines:
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise Exception('Could not retrieve version number')


setuptools.setup(
    name="guardcam",
    version=get_version(),
    author="Milogav",
    description="Visual movement detector in python with remote notifications",
    url='https://github.com/Milogav/GuardCam',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',

    install_requires=[get_requirements()],
    entry_points={
            "console_scripts": [
                "guardcam-start-detector = apps.start_detector:main",
            ]
        }
)
